import time
from buyable_molecules.create_postgres_db.funcs.db_setup import connect_to_db
from rdkit import DataStructs

def substructure_search(smarts, vendors=[], cursor=None, table='building_blocks', mols='mols', fps='fps', print_cmd=False):
    if cursor is None:
        cursor = connect_to_db().cursor()

    cmd = f"select {table}.*, bfp_to_binary_text({fps}.rdkit_fp) from {table} " \
          f"inner join {fps} on {table}.id={fps}.id " \
          f"where {table}.id in " \
          f"(SELECT {mols}.id from {mols} " \
          f"where {mols}.m@>'{smarts}'::qmol) " \

    num_vendors = len(vendors)
    if num_vendors != 0:
        cmd += " and ("

    for i, vendor in enumerate(vendors):
        cmd += f"{vendor}_id is not NULL"

        if i != num_vendors - 1:
            cmd += " or "
        else:
            cmd += ")"

    if print_cmd == True:
        print(cmd)

    cursor.execute(cmd)
    results = cursor.fetchall()

    new_results = []
    for i, result in enumerate(results):
        fp = DataStructs.CreateFromBinaryText(bytes(result[-1]))
        new_results.append(list(result[0:-1]) + [fp])

    return new_results

def substructure_search_single_vendor(smarts, vendor, cursor=None, table='building_blocks', mols='mols', fps='fps', print_cmd=False):
    if cursor is None:
        cursor = connect_to_db().cursor()

    cmd = f"SELECT {table}.*, bfp_to_binary_text({fps}.rdkit_fp) FROM " \
                f"(select {mols}.* from {table} " \
                f"inner join {mols} on {table}.id={mols}.id " \
                f"where {table}.{vendor}_id is not NULL) sq_mol " \
          f"inner join {fps} on {fps}.id=sq_mol.id " \
          f"inner join {table} on {table}.id=sq_mol.id " \
          f"where sq_mol.m@>'{smarts}'::qmol"

    if print_cmd == True:
        print(cmd)

    cursor.execute(cmd)
    results = cursor.fetchall()

    new_results = []
    for i, result in enumerate(results):
        fp = DataStructs.CreateFromBinaryText(bytes(result[-1]))
        new_results.append(list(result[0:-1]) + [fp])

    return new_results


if __name__ == "__main__":
    t0 = time.time()
    result = sq_substructure_search('CC(=O)[OH]', vendor='sigma')
    print(result)
    print(f"{round(time.time()-t0,2)} seconds")





import pandas as pd
from rdkit import Chem
from rdkit.Chem.SaltRemover import SaltRemover
remover = SaltRemover()
from rdkit import RDLogger
lg = RDLogger.logger()
lg.setLevel(RDLogger.ERROR)

def convert_to_rdkit(smi):
    try:
        mol = Chem.MolFromSmiles(smi)
        mol = remove_salt(mol)
        new_smi = Chem.MolToSmiles(mol)
        return new_smi
    except:
        print(f"{smi} not accepted by rdkit")
        return None

def remove_salt(mol):
    try:
        mol = remover.StripMol(mol, dontRemoveEverything=True)
    except:
        pass
    return mol

def remove_duplicate_smiles(df):
    df = df.drop_duplicates(subset=['SMILES'])
    return df

def create_mol_col(smi):
    try:
        return Chem.MolFromSmiles(smi)
    except:
        print(f'error with smi {smi}')
        return None

def get_inchi_key(mol):
    try:
        return Chem.MolToInchiKey(mol)
    except:
        return None

def load_mol_columns(df):
    df['mol'] = df['SMILES'].apply(create_mol_col)
    df = df.dropna(subset=['mol'])
    df['rdmol'] = df['mol'].apply(mol_binary)
    df = df.dropna(subset=['rdmol'])
    df['inchi_key'] = df['mol'].apply(get_inchi_key)
    df = df.dropna(subset=['inchi_key'])

    df = df.where(pd.notnull(df), None)

    return df

def nan_to_none(df):
    df = df.where(pd.notnull(df), None)
    return df

def create_merged_df(paths):
    list_dfs = load_dfs(paths)
    merged_df = merge_dfs(list_dfs)
    merged_df = nan_to_none(merged_df)
    return merged_df

def load_dfs(paths):
    print(f'loading {len(paths)} dataframes')
    list_dfs = []
    for path in paths:
        df = pd.read_csv(path, index_col=0)
        list_dfs.append(df)
    return list_dfs

def merge_dfs(list_dfs):
    print('merging dataframes')
    all_df = list_dfs.pop(0)
    for df in list_dfs:
        all_df = all_df.merge(df, on='SMILES', how='outer')
    return all_df


import os

USER = os.environ.get('PS_USER') or "postgres"
POSTGRES_PASSWORD = os.environ.get('PS_PASSWORD') or "rbc_postgres_1"
HOST = os.environ.get('PS_HOST') or "0.0.0.0"
PORT = os.environ.get('PS_PORT') or "5432"
DB_NAME = os.environ.get('PS_DB_NAME') or "molecules"


import psycopg2
import time


def create_mol_table(conn, cursor, table='building_blocks', mols='mols', id='id'):
    cursor.execute(f'drop table if exists {mols}')

    cmd = f"select * into {mols} from (select id,mol_from_smiles(smiles::cstring) m from {table}) tmp where m is not null;"
    cursor.execute(cmd)
    conn.commit()

    cmd = f"create index molidx on {mols} using gist(m)"
    cursor.execute(cmd)
    conn.commit()

    cmd = f"CREATE INDEX {mols}_idx_id ON {mols} ({id})"
    cursor.execute(cmd)
    conn.commit()

def create_fingerprints(conn, cursor, mols='mols', fps='fps', id_col='id'):
    from rdkit import Chem
    from rdkit.Chem import rdFingerprintGenerator
    from rdkit import DataStructs

    fp_list = []
    fp_gen = rdFingerprintGenerator.GetRDKitFPGenerator()
    # grab the molecules, we're pulling them out in their pickled form:
    print('getting molecules..')
    cursor.execute(f'select id, mol_send(m) from {mols}')
    results = cursor.fetchall()
    num_results = len(results)
    print(f"{num_results} molecules retrieved")
    print(f"creating fingerprints..")
    t0 = time.time()
    for i, (id, pkl) in enumerate(results):
        if pkl is None: continue
        m = Chem.Mol(pkl.tobytes())
        fp = fp_gen.GetFingerprint(m)
        fp_list.append((id, fp))

        if i % 10000 == 0 and i != 0:
            elapsed = round(time.time() - t0, 1)
            estimated_total_time = (elapsed / i) * num_results
            time_remaining = estimated_total_time - elapsed
            print(f"{i} added in {elapsed} seconds - estimated time remaining = {round(time_remaining / 60, 2)} mins")

    print('done')

    print('writing fps..')
    cursor.execute(f'drop table if exists {fps}')
    cursor.execute(f'create table {fps} (id int, rdkit_fp bfp)')
    cursor.executemany(f'insert into {fps} values (%s,bfp_from_binary_text(%s))',
                      [(x, DataStructs.BitVectToBinaryText(y)) for x, y in fp_list])
    conn.commit()
    print('done')

    print('making indexes..')
    cmd = f"CREATE INDEX {fps}_idx_id ON {fps}({id_col});"
    cursor.execute(cmd)
    conn.commit()
    #cmd = f"create index {fps}_rdfp_idx on {fps} using gist(rdkit_fp);"
    #cursor.execute(cmd)
    #conn.commit()
    print('done')





import psycopg2
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from buyable_molecules.config import POSTGRES_PASSWORD, USER, HOST, PORT, DB_NAME

def create_db(conn_string, db_name):
    conn = psycopg2.connect(conn_string)
    conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
    cursor = conn.cursor()
    cmd = f"create database {db_name}"
    cursor.execute(cmd)
    cursor.close()
    conn.close()
    print('database created')

def install_rd_extension(conn, cursor):
    cmd = "create extension if not exists rdkit;"
    cursor.execute(cmd)
    conn.commit()

def connect_to_db():
    try:
        conn = psycopg2.connect(
            f"user='{USER}' dbname='{DB_NAME}' host='{HOST}' password='{POSTGRES_PASSWORD}' port='{PORT}'")
    except:
        create_db(f"user='{USER}' host='{HOST}' password='{POSTGRES_PASSWORD}' port='{PORT}'", DB_NAME)
        conn = psycopg2.connect(
            f"user='{USER}' dbname='{DB_NAME}' host='{HOST}' password='{POSTGRES_PASSWORD}' port='{PORT}'")

    cursor = conn.cursor()
    install_rd_extension(conn, cursor)
    cursor.close()

    return conn

def table_columns(table, cursor):
    cursor.execute(f"Select * FROM {table} LIMIT 0")
    colnames = [desc[0] for desc in cursor.description]
    return colnames


from buyable_molecules.create_postgres_db.funcs.db_setup import connect_to_db
from buyable_molecules.urls import VENDORS


def smiles_lookup(smiles, cursor, table='building_blocks', vendors=[], print_cmd=False):
    cmd = f"select * from {table} where smiles='{smiles}'"


    num_vendors = len(vendors)
    if num_vendors != 0:
        cmd += " and ("

    for i, vendor in enumerate(vendors):
        cmd += f"{vendor}_id is not NULL"

        if i != num_vendors-1:
            cmd += " or "
        else:
            cmd += ")"

    if print_cmd == True:
        print(cmd)

    cursor.execute(cmd)
    result = cursor.fetchone()
    return result

def get_vendor_smiles(cursor, vendor, table='building_blocks'):
    cmd = f"select * from {table} where {vendor}_id is not NULL"
    cursor.execute(cmd)
    result = cursor.fetchall()
    return result

def test_smi_fp(cursor, smi, table='building_blocks', fps='fps'):
    cmd = f"select {table}.smiles, bfp_to_binary_text({fps}.rdkit_fp) " \
          f"from {table} " \
          f"inner join {fps} on {table}.id={fps}.id " \
          f"where smiles='{smi}'"
    cursor.execute(cmd)
    result = cursor.fetchall()
    return result

def vendor_rows(cursor, vendor, table='building_blocks'):
    cmd = f"select count(*) from {table} where {vendor}_id is not NULL"
    cursor.execute(cmd)
    result = cursor.fetchone()

    return result

def get_vendor_counts(cursor, vendors=VENDORS, table='building_blocks'):
    for v in VENDORS:
        num = vendor_rows(cursor, v)
        print(f"{v} - {num[0]}")

if __name__ == '__main__':
    conn = connect_to_db()
    cursor = conn.cursor()
    result = smiles_lookup('CCCCO', cursor)
    print(result)

    get_vendor_counts(cursor)




from buyable_molecules.create_csvs.funcs.download_and_process import download_and_process, download_and_process_lifechem, remove_file
from buyable_molecules.create_csvs.funcs import dataframe_functions
from pathlib import Path
from buyable_molecules import urls

SAVE_FOLDER = str(Path(__file__).parents[1]) + '/save_folder'
COLUMNS = ['SMILES', 'mcule_id', 'sigma_id', 'molport_id', 'zinc_id', 'apollo_id', 'flurochem_id']

def create_buyable_csv():
    mcule = download_and_process(urls.MCULE_URL, ["SMILES", "mcule_id"], None, '\t', f'{SAVE_FOLDER}/mcule.csv', download=True, unzip=True)
    sigma = download_and_process(urls.SIGMA_URL, ["SMILES", "sigma_id"], None, ' ', f'{SAVE_FOLDER}/sigma.csv', download=True, unzip=False)
    molport = download_and_process(urls.MOLPORT_URL, ["SMILES", "molport_id"], None, '\t', f'{SAVE_FOLDER}/molport.csv', download=True, unzip=False)
    zinc = download_and_process(urls.ZINC_IN_STOCK_URL, ["SMILES", "zinc_id", 'data_type', 'supplier', 'code'],
                         None, '\t', f'{SAVE_FOLDER }/zinc.csv', download=True, unzip=False)
    apollo = download_and_process(urls.APOLLO_BB_URL, ["SMILES", "apollo_id"], None, ' ', f'{SAVE_FOLDER}/apollo.csv',
                                 download=True, unzip=False)
    fluro = download_and_process(urls.FLUROCHEM_URL, ["SMILES", "flurochem_id"], None, ' ', f'{SAVE_FOLDER}/fluro.csv',
                                  download=True, unzip=False)

    alfa = download_and_process(urls.ALFA_URL, ["SMILES", "alfa_id"], None, ' ', f'{SAVE_FOLDER}/alfa.csv',
                                  download=True, unzip=False)

    lifechem = download_and_process_lifechem(urls.LIFECHEM_URL, f'{SAVE_FOLDER}/lifechem.csv')


    merged_df = dataframe_functions.create_merged_df([mcule, sigma, molport, zinc, apollo, fluro, alfa, lifechem])

    merged_df = merged_df[COLUMNS]
    merged_df.to_csv(f"{SAVE_FOLDER}/final.csv")

    remove_file(mcule)
    remove_file(sigma)
    remove_file(molport)
    remove_file(zinc)
    remove_file(apollo)
    remove_file(fluro)
    remove_file(alfa)
    remove_file(lifechem)

if __name__ == "__main__":

    #print('creating buyable csv..')
    #create_buyable_csv()
    print('..done')







MCULE_URL = "https://mcule.s3.amazonaws.com/database/mcule_purchasable_building_blocks_210714.smi.gz"
SIGMA_URL = 'http://files.docking.org/catalogs/source/sialbb.src.txt'
MOLPORT_URL = 'http://files.docking.org/catalogs/source/molportbb.src.txt'
ZINC_IN_STOCK_URL = 'http://files.docking.org/bb/current/In-Stock.txt'
APOLLO_BB_URL = "http://files.docking.org/catalogs/source/apollobb.src.txt"
FLUROCHEM_URL = "http://files.docking.org/catalogs/source/fluorochem.src.txt"
ALFA_URL = "http://files.docking.org/catalogs/source/alfa.src.txt" # https://www.alfa.com/en/catalog/H50684/
LIFECHEM_URL = "https://www.dropbox.com/s/d0hudapz6q12eas/LC_Stock_Building_Blocks.sdf?dl=1" #


VENDORS = ['mcule', 'sigma', 'molport', 'zinc', 'apollo', 'flurochem', 'alfa', 'lifechem']




import pandas as pd
from buyable_molecules.create_postgres_db.funcs import db_setup, create_building_blocks_table, create_mol_and_fp_tables
from buyable_molecules.create_csvs.funcs.download_and_process import remove_file
from buyable_molecules.create_csvs.create_buyable_csv import COLUMNS, SAVE_FOLDER

def create_db():
    print("creating db if it does not already exist")
    conn = db_setup.connect_to_db()
    cursor = conn.cursor()
    db_setup.install_rd_extension(conn, cursor)
    cursor.close()
    conn.close()

def create_building_blocks(remove_csv_after=True):
    print('Creating building_blocks table from df - this will take some time..')
    dtypes = {k: str for k in COLUMNS}
    df = pd.read_csv(f"{SAVE_FOLDER}/final.csv", index_col=0, dtype=dtypes)
    df.rename(columns={"SMILES": "smiles"}, inplace=True)
    print(df.head())

    conn = db_setup.connect_to_db()
    cursor = conn.cursor()

    create_building_blocks_table.sqlalchemy_data_to_table(df)
    create_building_blocks_table.create_primary_key(conn, cursor)

    cursor.close()
    conn.close()

    if remove_csv_after == True:
        remove_file(f"{SAVE_FOLDER}/final.csv")

def create_mols():
    print('Creating mols and fps tables - this will take some time..')
    conn = db_setup.connect_to_db()
    cursor = conn.cursor()

    create_mol_and_fp_tables.create_mol_table(conn, cursor)
    create_mol_and_fp_tables.create_fingerprints(conn, cursor)

    cursor.close()
    conn.close()

def make_db():
    create_db()
    create_building_blocks()
    create_mols()

if __name__ == '__main__':
    make_db()




from setuptools import setup, find_packages
from buyable_molecules import __version__

setup(
  name = 'buyable_molecules',
  packages = find_packages(),
  include_package_data=True,
  version = __version__,
  author = 'William Finnigan',
  author_email = 'wjafinnigan@gmail.com',
)




import pandas as pd
import numpy as np
import urllib.request
import gzip
import shutil
import os
from rdkit import Chem
from rdkit.Chem.SaltRemover import SaltRemover
from rdkit.Chem import PandasTools
remover = SaltRemover()
from rdkit import RDLogger
lg = RDLogger.logger()
lg.setLevel(RDLogger.ERROR)

from buyable_molecules.create_csvs.funcs import dataframe_functions

def load_data(path, cols, sep, header=None):
    print(f'Load path: {path}')
    data = pd.read_csv(path, sep=sep, header=header, names=cols)
    print(data.head())

    return data

def load_sdf(path):
    print(f'Load path: {path}')
    data = PandasTools.LoadSDF(path, smilesName='SMILES', molColName='Molecule',
                               includeFingerprints=False, removeHs=False, strictParsing=True)
    print(data.head())
    return data

def process_lifechem_sdf(df):
    def get_rdkit_smiles(mol):
        try:
            return Chem.MolToSmiles(mol)
        except:
            return None

    new_df = pd.DataFrame()
    df['Molecule'] = df['Molecule'].apply(dataframe_functions.remove_salt)
    new_df['SMILES'] = df['Molecule'].apply(get_rdkit_smiles)

    prices = []
    for index, row in df.iterrows():
        price = f"{row.get('Price_1', '')}$/{row.get('Pack_1', '')}"
        prices.append(price)
    new_df['lifechem_price'] = prices

    new_df['lifechem_id'] = df['IDNUMBER']

    new_df["SMILES"].replace('', np.nan, inplace=True)
    new_df = new_df.dropna(subset=['SMILES'])
    new_df = new_df.drop_duplicates()
    new_df = dataframe_functions.remove_duplicate_smiles(new_df)
    return new_df

def process_df_smiles(df_smiles):
    print('Data loaded - processing smiles..')
    print(f"Initial number of rows = {len(df_smiles.index)}")
    df_smiles["SMILES"] = df_smiles["SMILES"].apply(dataframe_functions.convert_to_rdkit)
    df_smiles["SMILES"].replace('', np.nan, inplace=True)
    df_smiles = df_smiles.dropna(subset=['SMILES'])
    df_smiles = df_smiles.drop_duplicates()
    df_smiles = dataframe_functions.remove_duplicate_smiles(df_smiles)
    print(f"Number of rows after processing = {len(df_smiles.index)}")

    return df_smiles



def download_file(url, filename):
    print(f"Downloading file from {url}")
    urllib.request.urlretrieve(url, filename)
    return filename

def download_and_unzip(url, filename='zip_file.gz', unzipped='unzipped.txt'):
    download_file(url, filename)
    print(f"Unzipping..")
    with gzip.open(filename, 'rb') as f_in:
        with open(unzipped, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    return unzipped

def remove_file(path):
    if os.path.exists(path):
        os.remove(path)

def select_columns(df, columns):
    new_df = df[columns]
    return new_df

def download_and_process(url_or_path, columns, header, sep, save_to, download=True, unzip=True):
    if download == True:
        if unzip == True:
            filename = download_and_unzip(url_or_path, filename='zip_file.gz', unzipped='file_to_process.txt')
            remove_file('zip_file.gz')
        else:
            filename = download_file(url_or_path, 'file_to_process.txt')
    else:
        filename = url_or_path

    df = load_data(filename, columns, sep, header=header)

    if download == True:
        remove_file(filename)

    df = process_df_smiles(df)
    df.to_csv(save_to)

    return save_to

def download_and_process_lifechem(url, save_to):
    file = download_file(LIFECHEM, 'lifechem.sdf')
    df = load_sdf(file)
    remove_file(file)
    df = process_lifechem_sdf(df)
    df.to_csv(save_to)
    return save_to

if __name__ == "__main__":
    from buyable_molecules.urls import LIFECHEM
    file = download_file(LIFECHEM, 'lifechem.sdf')
    df = load_sdf(file)
    df = process_lifechem_sdf(df)



    print(df.head())



import sqlalchemy

def sqlalchemy_data_to_table(df, table='building_blocks'):
    print("data to postgress..")
    engine = sqlalchemy.create_engine('postgresql://postgres:rbc_postgres_1@localhost:5432/molecules')
    df.to_sql(table, con=engine, index_label='id', if_exists='replace')
    print('done')

def create_primary_key(conn, cursor, table='building_blocks', id_col='id'):
    print('Create primary key and index..')
    cmd = f"alter table {table} add primary key ({id_col});"
    cursor.execute(cmd)
    conn.commit()
    cmd = f"CREATE INDEX {table}_idx_id ON {table} ({id_col});"
    cursor.execute(cmd)
    conn.commit()
    print('done')

def create_smiles_index(conn, cursor, table='building_blocks', smi_col='smiles'):
    cmd = f"CREATE INDEX {table}_idx_smiles ON {table} ({smi_col});"
    cursor.execute(cmd)
    conn.commit()
    print('done')

if __name__ == '__main__':
    from buyable_molecules.create_postgres_db.funcs import db_setup
    conn = db_setup.connect_to_db()
    cursor = conn.cursor()

    create_smiles_index(conn, cursor)

